# talKZ - (Kalfa and Zack)'s talk implementation
# Copyright (C) 2002:
# 	Cosimo Alfarano      <kalfa@cs.unibo.it>
# 	Stefano Zacchiroli   <zack@cs.unibo.it>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


from curses.ascii import isblank
import debugger


debug = False


def eraseChar(string):
    """ erase the last char of a string; return the new string """

    return string[:-1]

def eraseWord(string):
    """ erase the last word of a string acting like CTRL-W in shells or emacs;
    return the new string """

    i = len(string) - 1
    trim = 0
    while i >= 0:   # delete trailing blanks
        if isblank(string[i]):
            i -= 1
            trim += 1
        else:
            break
    if i>=0:
        assert not isblank(string[i])
    while i >= 0:   # delete backward until first blank
        if isblank(string[i]):
            break
        else:
            i -= 1
            trim += 1
    if trim > 0:    # trim string
        return string[:-trim]
    else:
        return string

def eraseLine(string):
    """ return an empty string """

    return ""


class UserLogger(object):
    """ per user logger, used by Logger """

    def __init__(self, user, local=False):
        """
        UserLogger(user [,local=bool])
            'user'      user name, used to qualify messages
            'local'     true if this is the local user, used to apply
                        a particular format to local user log
        """

        self.__user = user
        self.__local = local
        self.resetLine()

    def resetLine(self):
        self.__curLine = "%s > " % self.__user
        if self.__local:    # special format for local user
            self.__curLine = "@ " + self.__curLine

    def eraseChar(self):
        self.__curLine = eraseChar(self.__curLine)

    def eraseWord(self):
        self.__curLine = eraseWord(self.__curLine)

    def eraseLine(self):
        self.__curLine = eraseLine(self.__curLine)

    def __validateChar(self, char):
        if type(char) != str:
            raise TypeError("'char' must be a character")
        if len(char) != 1:
            raise ValueError("'char' must have length 1")

    def __validateString(self, string):
        if type(string) != str:
            raise TypeError("'string' must be a string")

    def logChar(self, char):
        self.__validateChar(char)
        self.__curLine += char

    def logString(self, string):
        self.__validateString(string)
        for c in string:
            self.logChar(c)

    def getLine(self):
        return self.__curLine


class Logger(object):
    """ Talk session logger: keep and save the log for a talk session between
    two users in a line buffered fashion.

    Logger(lname, rname, [logfile, [mode, [logmagics]]])
    'lname'     name of the local user
    'rname'     name of the remote user
    'logfile'   name of the logfile
    'mode'      opening mode for the logfile, can be 'w' for overwrite mode or
                'a' for append mode; default is 'w'
    'logmacis'  whether to log special characters like backspace or not. If
                this feature is on (logmagics == True) log file record also
                backspace and other control characters, otherwise special
                character are "executed", for example if the user "zack" enters
                this key sequence: 'f' 'u' 'b' 'a' 'x' <BACKSPACE> 'r' you will
                have these possible logfile:

                    logmagics == True   ->  'zack > fubax^Hr'
                    logmagics == False  ->  'zack > fubar'

                Default is off (logmagics == False)

    Public methods:
        logChar{L,R}        log a 'normal' char
        logEraseChar{L,R}   log a backspace char
        logEraseWord{L,R}   log an erase word char
        logEraseLine{L,R}   log an erase line char
        logEol{L,R}         log an end of line, note that this cause the last
                            log line to be committed on the logfile

    Please note that this interface is based on python characters (i.e. string
    of length 1), _not_ on C-like characters (i.e. integers)
    """

    def __init__(self, lname, rname, logfile="talkz.log", mode="w",
                logmagics=False):
        self.__CTRL_H = '\x08'  # CTRL-H (i.e. backspace)
        self.__CTRL_W = '\x17'  # CTRL-W (i.e. erase word)
        self.__CTRL_U = '\x15'  # CTRL-U (i.e. erase line)
        self.__EOL = '\n'       # End of line
        self.__logmagics = logmagics
        self.__loggers = {}
        self.__loggers["local"] = UserLogger(lname, local=True)
        self.__loggers["remote"] = UserLogger(rname)
        if type(mode) != str:
            raise TypeError("'mode' must be a character")
        if (mode != 'w') and (mode != 'a'):
            raise ValueError("'mode' must be one of 'a', 'w'")
        self.__logfile = open(logfile, mode)
        self.__desc =   \
            "<logfile=%s, mode=%s, lname=%s, rname=%s, logmagics=%d>" % \
            (logfile, mode, lname, rname, logmagics)
        if debug:
            self.__debugger = debugger.Debugger(fname="LOGGER.PY.LOG")
            self.__debugger.log("Logger started!")

    def __repr__(self):
        return self.__desc

    def stop(self):
        self.logEolL()
        self.logEolR()
        self.__logfile.close()

    def __del__(self):
        self.stop()

    def __logChar(self, char, where):
        self.__loggers[where].logChar(char)

    def logCharL(self, char):
        self.__logChar(char, "local")
        if debug:
            self.__debugger.log("logging local char")

    def logCharR(self, char):
        self.__logChar(char, "remote")
        if debug:
            self.__debugger.log("logging remote char")

    def __logEraseChar(self, where):
        logger = self.__loggers[where]
        if self.__logmagics:
            logger.logChar(self.__CTRL_H)
        else:   # don't log magic chars
            logger.eraseChar()

    def logEraseCharL(self):
        self.__logEraseChar("local")
        if debug:
            self.__debugger.log("logging local erase char")

    def logEraseCharR(self):
        self.__logEraseChar("remote")
        if debug:
            self.__debugger.log("logging remote erase char")

    def __logEraseWord(self, where):
        logger = self.__loggers[where]
        if self.__logmagics:
            logger.logChar(self.__CTRL_W)
        else:   # don't log magic chars
            logger.eraseWord()

    def logEraseWordL(self):
        self.__logEraseWord("local")
        if debug:
            self.__debugger.log("logging local erase word")

    def logEraseWordR(self):
        self.__logEraseWord("remote")
        if debug:
            self.__debugger.log("logging remote erase word")

    def __logEraseLine(self, where):
        logger = self.__loggers[where]
        if self.__logmagics:
            logger.logChar(self.__CTRL_U)
        else:   # don't log magic chars
            logger.eraseLine()

    def logEraseLineL(self):
        self.__logEraseLine("local")
        if debug:
            self.__debugger.log("logging local erase line")

    def logEraseLineR(self):
        self.__logEraseLine("remote")
        if debug:
            self.__debugger.log("logging remote erase line")

    def __logEol(self, where):
        logger = self.__loggers[where]
        self.__logfile.write(logger.getLine() + self.__EOL)
        self.__logfile.flush()
        logger.resetLine()

    def logEolL(self):
        self.__logEol("local")
        if debug:
            self.__debugger.log("logging local end of line")

    def logEolR(self):
        self.__logEol("remote")
        if debug:
            self.__debugger.log("logging remote end of line")

