# talKZ - (Kalfa and Zack)'s talk implementation
# Copyright (C) 2002:
# 	Cosimo Alfarano      <kalfa@cs.unibo.it>
# 	Stefano Zacchiroli   <zack@cs.unibo.it>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

""" Messages exchanged between client and server """

from os import getlogin
from os import getpid
from re import split
from socket import inet_ntoa, inet_aton
import re
import socket

from classes import *
import kzstruct


debug = False    # debugging on/off

TALK_VERSION = 1    # default talk version
TALKZ_VERSION = 165 # talkz version (== ord('K') + ord('Z'))
VERSIONS = [ TALK_VERSION, TALKZ_VERSION ]
CAPA = { 'tls': chr(166), 'none': chr(167) }
CAPA_LEN = 1

# request type
LEAVE_INVITE = 0
LOOK_UP = 1
DELETE = 2
ANNOUNCE = 3

# response type
SUCCESS = 0
NOT_HERE = 1
FAILED = 2
MACHINE_UNKNOWN = 3
PERMISSION_DENIED = 4
UNKNOWN_REQUEST = 5
BADVERSION = 6
BADADDR = 7
BADCTLADDR = 8


def string_of_reqtyp(typ):
    if typ == LEAVE_INVITE:
        str = "LEAVE_INVITE"
    elif typ == LOOK_UP:
        str = "LOOK_UP"
    elif typ == DELETE:
        str = "DELETE"
    elif typ == ANNOUNCE:
        str = "ANNOUNCE"
    else:
        raise ValueError("unknown request type: '%d'" % typ)
    return str

def string_of_answer(answer):
    if answer == SUCCESS:
        str = "SUCCESS"
    elif answer == NOT_HERE:
        str = "NOT_HERE"
    elif answer == FAILED:
        str = "FAILED"
    elif answer == MACHINE_UNKNOWN:
        str = "MACHINE_UNKNOWN"
    elif answer == PERMISSION_DENIED:
        str = "PERMISSION_DENIED"
    elif answer == UNKNOWN_REQUEST:
        str = "UNKNOWN_REQUEST"
    elif answer == BADVERSION:
        str = "BADVERSION"
    elif answer == BADADDR:
        str = "BADADDR"
    elif answer == BADCTLADDR:
        str = "BADCTLADDR"
    else:
        raise ValueError("unknown answer type: '%d'" % answer)
    return str


def combine(l1, l2):
    """ given a pair of lists return a list of pairs """
    return(map((lambda x,y: (x,y)), l1, l2))


def trunc_at_0(s):
    """ convert C-like strings to pyton-like strings (i.e. trunc a string at
    the first null byte encountered) """
    try:
        return s[:s.index('\x00')]
    except ValueError:
        return s


class Address(object):
    """
    Talk addresses, carry an address family, an ip address and a port.
    Internally, address value is ketp in network byte order (suitable to be
    used with inet_ntoa function) whether port and family are kept in host byte
    order

    Methods:
        setAddr(address)    sets a new addr value, address is a string
        getAddr()           return the current addr value, in binary form,
                            network byte order
        getStrAddr()        like getAddr() but return a human readable
                            representation
        setPort(port)       sets a new port value, port can be a string or an
                            integer
        getPort()           return the current port value as python
                            object
        setFamily(family)   sets a new family value, family is an integer
        getFamily()         return the current family value as an integer
        isEqual(addr)       equality predicate over Address objects
    """

    def __repr__(self):
        return "<%s, %d, %d>" % \
            (socket.inet_ntoa(self.getAddr()), self.getPort(),
            self.getFamily())

    def __init__(self, addr, port, family=socket.AF_INET):
        self.__address = inet_aton(addr)
        self.__port = int(port)
        self.__family = family

    def setAddr(self, addr):
        """ setAddr(address) sets a new addr value, 'address' is a string """
        self.__address = inet_aton(addr)

    def getAddr(self):
        """ return the current address in network byte order (suitable for
        socket.inet_ntoa) """
        return self.__address

    def getStrAddr(self):
        """ return a human readable representation of self.__address """

        return inet_ntoa(self.__address)

    def setPort(self, port):
        """ setPort(port) sets a new port value, 'port' is a string """
        self.__port = int(port)

    def getPort(self):
        """ return the current port value in host byte order """
        return self.__port

    def setFamily(self, family):
        """ setFamily(family) sets a new family value, 'family' is an integer
        (e.g. socket.AF_INET) """
        self.__family = family

    def getFamily(self):
        """ return the current family value in host byte order """
        return self.__family

    def isEqual(self, addr):
        """ equality predicate over addresses """
        return (    \
            (self.getAddr() == addr.getAddr())  and \
            (self.getPort() == addr.getPort())  and \
            (self.getFamily() == addr.getFamily()))


class Message(object):
    """ Abstraction over messages exchanged between client and server """

    def pack(self):
        """
        Return a binary representation of the message suitable to be
        transmitted over the network between client and server
        """

        raise AbstractMethod("pack")

# Problematic: python toplevel use this function to show computation values, so
# this definition will mess up python toplevel debugging ...
#     def __repr__(self):
#         """ Wrapper around self.pack """
# 
#         return self.pack()


class Request(Message):
    """
    Abstraction over requests (i.e. messages from client to server)

    Methods:
        getType()       return the request type
        setType(typ)
        getId()         return the request id_num
        setId(id)
        getAddr()       return the request address (used to due effective
                        talking)
        setAddr(a)
        getCtlAddr()    return the request control address (used to carry on
                        the talk protocol with the talk servers)
        setCtlAddr(a)
        getPid()        return the request pid
        setPid(pid)
        getCallerName() return the caller login name
        setCallerName(name)
        getCalleeName() return the callee login name
        setCalleeName(name)
        getCalleeTty()  return the callee tty
        setCalleeTty(tty)
        getVersion()    return the request version
        setVersion(vers)
        isEqual(req)    equality predicate over requests
    """

    def __init__(self, typ, id_num, addr, ctl_addr,
                pid, l_name, r_name, r_tty, vers=TALK_VERSION):
        """
        Request(typ, id_num, addr, ctl_addr, pid, l_name, r_name, r_tty, vers)

        - typ is the type of answer, one of: LEAVE_INVITE, LOOK_UP, DELETE,
          ANNOUNCE
        - id_num is the message id
        - addr and ctl_addr are Address object (see above), "addr" is the
          client tcp port address, "ctl_addr" is the client udp port
        - pid is an integer representing the caller's process id
        - l_name is a string representing the caller's login name
        - r_name is a string representing the callee's login name
        - r_tty is a string representing the callee's tty
        - vers is an integer representing the protocol version, this argument
          is optional and defaults to "TALK_VERSION"
        """
        assert isinstance(addr, Address) and isinstance(ctl_addr, Address)
        self._typ = typ
        self._id_num = id_num
        self._addr = addr
        self._ctl_addr = ctl_addr
        self._pid = pid
        self._l_name = l_name
        self._r_name = r_name
        self._r_tty = r_tty
        self._vers = vers

    def getType(self):
        return self._typ

    def setType(self, typ):
        if typ not in [LEAVE_INVITE, LOOK_UP, DELETE, ANNOUNCE]:
            raise ValueError("Unknown type: '%d'" % typ)
        self._typ = typ

    def getId(self):
        return self._id_num

    def setId(self, id):
        if type(id) != int:
            raise TypeError("id must be an integer")
        self._id_num = id

    def getAddr(self):
        return self._addr

    def setAddr(self, addr):
        if not isinstance(addr, Address):
            raise TypeError("addr must be an Address object")
        self._addr = addr

    def getCtlAddr(self):
        return self._ctl_addr

    def setCtlAddr(self, ctl_addr):
        if not isinstance(ctl_addr, Address):
            raise TypeError("ctl_addr must be an Address object")
        self._ctl_addr = ctl_addr

    def getPid(self):
        return self._pid

    def setPid(self, pid):
        if not type(pid) == int:
            raise TypeError("'pid' must be an integer")
        self._pid = pid

    def getCallerName(self):
        return self._l_name

    def setCallerName(self, caller_name):
        if not type(caller_name) == str:
            raise TypeError("'caller_name' must be a string")
        self._l_name = caller_name

    def getCalleeName(self):
        return self._r_name

    def setCalleeName(self, callee_name):
        if not type(callee_name) == str:
            raise TypeError("'callee_name' must be a string")
        self._r_name = callee_name

    def getCalleeTty(self):
        return self._r_tty

    def setCalleeTty(self, callee_tty):
        if not type(callee_tty) == str:
            raise TypeError("'callee_tty' must be a string")
        self._r_tty = callee_tty

    def getVersion(self):
        return self._vers

    def setVersion(self, version):
        if type(version) != int:
            raise TypeError("'version' must be an integer")
        self._vers = version

    def isEqual(self, req):
        """ equality predicate over requests """
        return (    \
            (self.getType() == req.getType()) and \
            (self.getId() == req.getId())   and \
            (self.getAddr().isEqual(req.getAddr())) and \
            (self.getCtlAddr().isEqual(req.getCtlAddr()))   and \
            (self.getPid() == req.getPid()) and \
            (self.getCallerName() == req.getCallerName())   and \
            (self.getCalleeName() == req.getCalleeName())   and \
            (self.getCalleeTty() == req.getCalleeTty()) and \
            (self.getVersion() == req.getVersion()))


class Response(Message):
    """
    Abstraction over requests (i.e. messages from client to server)

    Methods:
        getType()       return the response type
        setType(typ)    set the response type
        getAnswer()     return the response answer value
        setAnswer(ans)  set the response answer value
        getId()         return the response id_num
        setId(id)       set the response id_num
        getAddr()       return the response address
        setAddr(addr)   set the response address
        getVersion()    return the response version
        setVersion(v)   set the response version
        isEqual(req)    equality predicate over responses
    """

    def __init__(self, typ, answer, id_num, addr, vers=TALK_VERSION):
        """
        Response(typ, answer, id_num, addr, vers)

        - typ is the type of request which we are responding, see Request
        - answer is the type of answer, one of: SUCCESS, NOT_HERE, FAILED,
          MACHINE_UNKNOWN, PERMISSION_DENIED, UNKNOWN_REQUEST, BADVERSION,
          BADADDR, BADCTLADDR
        - id_num is the message id
        - addr is an Address object (see above) which is the address used to
          estabilish a conversation
        - vers is an integer representing the protocol version, this argument
          is optional and defaults to "TALK_VERSION"
        """
        assert isinstance(addr, Address)
        self._typ = typ
        self._answer = answer
        self._id_num = id_num
        self._addr = addr
        self._vers = vers

    def getType(self):
        return self._typ

    def setType(self, typ):
        if typ not in [LEAVE_INVITE, LOOK_UP, DELETE, ANNOUNCE]:
            raise ValueError("Unknown type: '%d'" % typ)
        self._typ = typ

    def getAnswer(self):
        return self._answer

    def setAnswer(self, answer):
        if answer not in [SUCCESS, NOT_HERE, FAILED, MACHINE_UNKNOWN,
                    PERMISSION_DENIED, UNKNOWN_REQUEST, BADVERSION,
                    BADADDR, BADCTLADDR]:
            raise ValueError("Unknown answer: '%d'" % answer)
        self._answer = answer

    def getId(self):
        return self._id_num

    def setId(self, id):
        if type(id) != int:
            raise TypeError("id must be an integer")
        self._id_num = id

    def getAddr(self):
        return self._addr

    def setAddr(self, addr):
        if not isinstance(addr, Address):
            raise TypeError("addr must be an Address object")
        self._addr = addr

    def getVersion(self):
        return self._vers

    def setVersion(self, version):
        self._vers = version

    def isEqual(self, res):
        """ equality predicate over responses """
        return (    \
            (self.getType() == res.getType()) and \
            (self.getAnswer() == res.getAnswer())   and \
            (self.getId() == res.getId())   and \
            (self.getAddr().isEqual(res.getAddr())) and \
            (self.getVersion() == res.getVersion()))


class Bsd(object):
    """
    BSD talk protocol specific data

    Static methods:
        getReqSize()    return size of a serialized request object
        getResSize()    return size of a serialized response object

    Reference C data structures for packing:

    #define	NAME_SIZE	12
    #define	TTY_SIZE	16

    struct talk_addr {
      u_int16_t ta_family;
      u_int16_t ta_port;
      u_int32_t ta_addr;
      u_int32_t ta_junk1;
      u_int32_t ta_junk2;
    };
    
    typedef struct {
        u_int8_t    vers;
        u_int8_t    type;
        u_int8_t    answer;
        u_int8_t    pad;
        u_int32_t    id_num;
        struct talk_addr addr;
        struct talk_addr ctl_addr;
        u_int32_t    pid;
        char        l_name[NAME_SIZE];
        char        r_name[NAME_SIZE];
        char        r_tty[TTY_SIZE];
    } CTL_MSG;
    
    typedef struct {
        u_int8_t    vers;
        u_int8_t    type;
        u_int8_t    answer;
        u_int8_t    pad;
        u_int32_t   id_num;
        struct    talk_addr addr;
    } CTL_RESPONSE;
    """

    name_size = 12
    tty_size = 16
    _fmts = {}
    _fmts["talk_addr"] = "!H !H 4s 8x"  # N.B. Address.getAddr() return address
                                        # already in network byte order
    _fmts["ctl_msg"] = "B B 2x !I %s %s !I %ds %ds %ds" % \
                        (_fmts["talk_addr"], _fmts["talk_addr"],
                         name_size, name_size, tty_size)
    _fmts["ctl_response"] = "B B B x !I %s" % (_fmts["talk_addr"])


    def getReqSize():
        """ return the expected size of a request object serialization """

        return kzstruct.calcsize(Bsd._fmts["ctl_msg"])
    getReqSize = staticmethod(getReqSize)

    def getResSize():
        """ return the expected size of a response object serialization """

        return kzstruct.calcsize(Bsd._fmts["ctl_response"])
    getResSize = staticmethod(getResSize)


class SunOs(object):
    """
    SunOs talk protocol specific data

    Static methods:
        getReqSize()    return size of a serialized request object
        getResSize()    return size of a serialized response object

    Reference C data structures for packing:

    struct osockaddr
      {
        unsigned short int sa_family;
        unsigned char sa_data[14];
      };

    typedef struct {
            u_char  type;
    #define NAME_SIZE_S       9
            char    l_name[NAME_SIZE_S];
            char    r_name[NAME_SIZE_S];
            u_char  pad;
            u_int32_t id_num;
            int32_t pid;
            char    r_tty[TTY_SIZE];
            struct  osockaddr addr;
            struct  osockaddr ctl_addr;
    } CTL_MSG_S;

    typedef struct {
            u_char  type;
            u_char  answer;
            u_short  pad;
            u_long id_num;
            struct  osockaddr addr;
    } CTL_RESPONSE_S;
    """

        # TODO see where to put a network order ("!") flag!!!
    name_size_s = 9
    tty_size = 16
    _fmts = {}
    _fmts["osockaddr"] = "H 14B"
    _fmts["ctl_msg"] = "B %ds %ds x I i %ds %s %s" % \
                        (name_size_s, name_size_s, tty_size,
                        _fmts["osockaddr"], _fmts["osockaddr"])
    _fmts["ctl_response"] = "B B 2x L %s" % (_fmts["osockaddr"])

    def getReqSize():
        """ return the expected size of a request object serialization """

        return kzstruct.calcsize(SunOs._fmts["ctl_msg"])
    getReqSize = staticmethod(getReqSize)

    def getResSize():
        """ return the expected size of a response object serialization """

        return kzstruct.calcsize(SunOs._fmts["ctl_response"])
    getResSize = staticmethod(getResSize)


class BsdRequest(Bsd, Request):
    """ BSD protocol (linux default) requests """

    def __repr__(self):
        return  \
"<bsdRequest>\n\
Version:    %d\n\
Type:       %s\n\
IdNum:      %d\n\
Addr:       %s\n\
CtlAddr:    %s\n\
Pid:        %d\n\
LName:      %s\n\
RName:      %s\n\
RTty:       %s\n\
</bsdRequest>\n" %   \
        (self._vers, string_of_reqtyp(self._typ), self._id_num, `self._addr`,
        `self._ctl_addr`, self._pid, self._l_name, self._r_name, self._r_tty)

    def pack(self):
        if debug: print "Packing a BsdRequest ..."
        return kzstruct.pack(self._fmts["ctl_msg"],
                [self._vers, self._typ, self._id_num, self._addr.getFamily(),
                self._addr.getPort(), self._addr.getAddr(),
                self._ctl_addr.getFamily(), self._ctl_addr.getPort(),
                self._ctl_addr.getAddr(), self._pid, self._l_name,
                self._r_name, self._r_tty])


class BsdResponse(Bsd, Response):
    """ BSD protocol (linux default) responses """

    def __repr__(self):
        return  \
"<bsdResponse>\n\
Version:    %d\n\
Type:       %s\n\
Answer:     %s\n\
IdNum:      %d\n\
Addr:       %s\n\
</bsdResponse>\n" %   \
        (self._vers, string_of_reqtyp(self._typ),
        string_of_answer(self._answer), self._id_num, `self._addr`)

    def pack(self):
        if debug: print "Packing a BsdResponse ..."
        return kzstruct.pack(self._fmts["ctl_response"],
                [self._vers, self._typ, self._answer, self._id_num,
                self._addr.getFamily(), self._addr.getPort(),
                self._addr.getAddr()])


class SunOsRequest(SunOs, Request):
    """ SunOs protocol requests """

    def pack(self):
        # TODO
        raise NotImplemented("pack")


class SunOsResponse(SunOs, Response):
    """ SunOs protocol responses """

    def pack(self):
        # TODO
        raise NotImplemented("pack")


class BsdMessageFactory(Bsd):
    """
    Factory for bsd talk protocol messages, this class is not supposed to be
    used directly, use MessageFactory (see below) instead.

    Methods:
       buildRequest(...)    build a request given all needed args
       buildResponse(...)   build a response giben all needed args
       buildRequestFromBinary(rawdata)
                            build a request unpacking the given rawdata
       buildResponseFromBinary(rawdata)
                            build a response unpacking the given rawdata
    """

    def buildRequest(self, addr, ctl_addr, r_name, r_tty, vers, \
                    typ, id_num, pid, l_name):
        return BsdRequest(addr=addr, ctl_addr=ctl_addr, r_name=r_name, \
                            r_tty=r_tty, vers=vers, typ=typ, id_num=id_num, \
                            pid=pid, l_name=l_name)

    def buildResponse(self, typ, answer, id_num, addr, vers):
        if addr == None:
            addr = Address("0.0.0.0", 0)
        return BsdResponse(typ=typ, answer=answer, id_num=id_num, \
                            addr=addr, vers=vers)

    def buildRequestFromBinary(self, rawdata):
        actual_len = len(rawdata)
        expected_len = kzstruct.calcsize(self._fmts["ctl_msg"])
        if actual_len != expected_len:
            raise ValueError(
                "size of binary data is %d, expected size should be %d" %   \
                (actual_len, expected_len))
        try:
            [vers, typ, id_num, addr_family, addr_port, addr_addr,
            ctl_addr_family, ctl_addr_port, ctl_addr_addr, pid, l_name,
            r_name, r_tty] = \
                kzstruct.unpack(self._fmts["ctl_msg"], rawdata)
        except kzstruct.Error, error:
            raise ValueError(error.getMsg())
        (l_name, r_name, r_tty) = \
            map ((lambda s: trunc_at_0(s)), (l_name, r_name, r_tty))
        addr = Address(inet_ntoa(addr_addr), addr_port, addr_family)
        ctl_addr = Address(inet_ntoa(ctl_addr_addr), ctl_addr_port,
                        ctl_addr_family)
        return BsdRequest(typ=typ, id_num=id_num, addr=addr, ctl_addr=ctl_addr,
                        pid=pid, l_name=l_name, r_name=r_name, r_tty=r_tty,
                        vers=vers)

    def buildResponseFromBinary(self, rawdata):
        actual_len = len(rawdata)
        expected_len = kzstruct.calcsize(self._fmts["ctl_response"])
        if actual_len != expected_len:
            raise ValueError(
                "size of binary data is %d, expected size should be %d" %   \
                (actual_len, expected_len))
        try:
            [vers, typ, answer, id_num, addr_family, addr_port, addr_addr] = \
                kzstruct.unpack(self._fmts["ctl_response"], rawdata)
        except kzstruct.Error, error:
            raise ValueError(error.getMsg())
        addr = Address(inet_ntoa(addr_addr), addr_port, addr_family)
        return BsdResponse(typ=typ, answer=answer, id_num=id_num, addr=addr,
                        vers=vers)


class SunOsMessageFactory(SunOs):
    """ factory for sunos talk protocol messages, this class is not supposed to
    be used directly, use MessageFactory (see below) instead
    
    NOT IMPLEMENTED
    """

    def __init__(self):
        # TODO
        raise NotImplemented("SunOsMessageFactory")

class MessageFactory(object):
    """
    factory used to build talk protocol messages

    Methods:
        getProtocol()       return the protocol for which messages are built
        setProtocol(proto)  set the protocol for which messages are built
        switchProtocol()    switch between "sunos" and "bsd" protocols
        getProtocolClass()  return the class (not an instance!!) corresponding
                            to the current protocol (i.e. if protocol == "bsd"
                            return the Bsd class)
        buildRequest(...)   create a request message
        buildResponse(...)  create a response message
        buildRequestFromBinary(rawdata)     create a request unpacking binary data
        buildResponseFromBinary(rawdata)    create a response unpackging binary data
    """

    def __setUpFactory(self):
        """ update (if needed) internal factory so that is is in sync with the
        self.__protocol current value """

        if self.__protocol == "bsd":
            self.__factory = BsdMessageFactory()
        elif self.__protocol == "sunos":
            self.__factory = SunOsMessageFactory()
        else:   # unsupported protocol
            raise "Unsupported protocol: '%s'" % (self.__protocol)

    def __init__(self, protocol="bsd"):
        """
        MessageFactory(protocol)
        where protocol can be one of 'bsd', 'sunos'
        """

        self.__protocol = protocol
        self.__setUpFactory()
    
    def getProtocol(self):
        """ return the current protocol, which correspond to the type of built
        messages """

        return self.__protocol

    def setProtocol(self, protocol):
        """ change the current protocol to a new value and change the internal
        factory accordingly if needed """

        if (protocol != "bsd") and (protocol != "sunos"):
            raise ValueError("'protocol' must be one of 'sunos', 'bsd'")
        if protocol != self.__protocol:
            self.__protocol = protocol
            self.__setUpFactory()

    def switchProtocol(self):
        """ switch the current protocol: if actually we are using the 'bsd'
        protocol change it to 'sunos' and viceversa """

        if self.__protocol == "bsd":
            newprotocol = "sunos"
        elif self.__protocol == "sunos":
            newprotocol = "bsd"
        else:
            raise "Inconsistent state: self.__protocol = '%s'" % (protocol)
        self.setProtocol(newprotocol)

    def getProtocolClass(self):
        if self.__protocol == "bsd":
            c = Bsd
        elif self.__protocol == "sunos":
            c = SunOs
        else:
            raise "Inconsistent state: self.__protocol = '%s'" % (protocol)
        return c

    def buildRequest(self, addr, ctl_addr, r_name, r_tty, vers=TALK_VERSION, \
                    typ=LOOK_UP, id_num=0, pid=None, l_name=None):
        """ create a request object using the current factory """

        if pid == None:
            pid = getpid()
        if l_name == None:
            l_name = getlogin()
        return self.__factory.buildRequest(addr=addr, ctl_addr=addr, \
                                    r_name=r_name, r_tty=r_tty, vers=vers, \
                                    typ=typ, id_num=id_num, pid=pid, \
                                    l_name=l_name)

    def buildResponse(self, addr=None, vers=TALK_VERSION, typ=LOOK_UP, \
                        id_num=0, answer=SUCCESS):
        """ create a response object using the current factory """
        return self.__factory.buildResponse(typ=typ, answer=answer, \
                                    id_num=id_num, addr=addr, vers=vers)

    def buildRequestFromBinary(self, rawdata):
        """ like buildRequest but unpack the given 'rawdata' argument to find
        out construction parameters to build a request """
        return self.__factory.buildRequestFromBinary(rawdata)

    def buildResponseFromBinary(self, rawdata):
        """ like buildResponse but unpack the given 'rawdata' argument to find
        out construction parameters to build a response """
        return self.__factory.buildResponseFromBinary(rawdata)

