# talKZ - (Kalfa and Zack)'s talk implementation
# Copyright (C) 2002:
# 	Cosimo Alfarano      <kalfa@cs.unibo.it>
# 	Stefano Zacchiroli   <zack@cs.unibo.it>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import curses
import curses.ascii
import sys
import termios

import debugger
import ui


debug = False


def isEol(char):
    """ predicate: check if 'char' is an end of line character or not """

    # TODO: do we need to consider other characters?
    if type(char) != str:
        raise TypeError("'char' must be a string of length 1")
    return (char == '\r') or (char == '\n')


class NCursesUI(ui.UserInterface):
    """
    Textual, ncurses-based user interface for talkz clients.
    The user interface is something like:

        |----------------------------|
        |                            |
        |      local window          |
        |                            |
        |-- status window (1 line) --|
        |                            |
        |      remote window         |
        |                            |
        |----------------------------|

    'local window' is used to echo chars written by the local user
    'remote window' is used do siplay chars received from remote user
    'status window' is a one-line window used to inform the local user about
    the talk status (e.g. "connecting to peer", "talking", ...)
    """

    def __init__(self, logger=None):
        """ 
        Get magic chars and initialize internal data structures.

        NCursesUI([logger])
            'logger'    if passed and different from 'None' is the logger used
                        to log talk session, see logger.py
        """

        if debug:
            self.__debugger = debugger.Debugger(fname="NCURSESUI.PY.LOG")

        self.__logger = logger

            # set up magic characters
        term_magics = self.__getTermMagicChars()
        magiccharsL = term_magics
        magiccharsR = term_magics
        self.setMagicCharsL(magiccharsL)
        self.setMagicCharsR(magiccharsR)
        self.__running = False

    def start(self):
        """ Start ncurses and draw windows """

        self.__boot()
        self.__buildWindows()
        self.__refreshAll() # draw user interface
        self.__running = True

    def isRunning(self):
        return self.__running

    def __getTermMagicChars(self):
        cc_index = 6
        ctrl_chars = termios.tcgetattr(sys.stdin.fileno())[cc_index]
        cerase = ord(ctrl_chars[termios.VERASE])
        werase = ord(ctrl_chars[termios.VWERASE])
        lerase = ord(ctrl_chars[termios.VKILL])
        return (cerase, werase, lerase)
    
    def __buildWindows(self):
        """ build 3 main windows (local, remote, status), set up their
        attributes and draw them (unless do_draw==False) """

            # compute sizes
        statusH = 1
        localH = (self.__LINES - statusH) / 2   # rounded down
        remoteH = (self.__LINES - statusH - localH) # rounded up
        statusW = self.__COLS
        localW = self.__COLS
        remoteW = self.__COLS

            # build windows
        lwin = curses.newwin(localH, localW, 0, 0)
        swin = curses.newwin(statusH, statusW, localH, 0)
        rwin = curses.newwin(remoteH, remoteW, localH + statusH, 0)
        lwin.idlok(1)   # enable scrolling for locale and remote windows
        lwin.scrollok(1)
        rwin.idlok(1)
        rwin.scrollok(1)
        lwin.keypad(1)  # catch easily kbd control sequences
        rwin.keypad(1)
        swin.attron(curses.A_REVERSE)   # revers video for status line

        self.__wins = {}    # windows dictionary
        self.__wins["local"] = lwin
        self.__wins["status"] = swin
        self.__wins["remote"] = rwin

    def __boot(self):
        """ curses initialization """

        stdscr = curses.initscr()
        curses.noecho()
        curses.cbreak()
#         curses.curs_set(1)
        (max_y, max_x) = stdscr.getmaxyx()
        self.__LINES = max_y
        self.__COLS = max_x

    def __refreshAll(self):
        """ refresh all windows efficiently """

        for w in self.__wins.values():
            w.noutrefresh()
        curses.doupdate()

    def __refresh(self, win):
        """ note: win here is a ncurses window, _not_ a window name used to
        index self.__wins """

        lwin = self.__wins["local"]
        win.refresh()
        if win != lwin: # refresh also local window so that cursor is there
            lwin.refresh()

    def refresh(self):
        """ public frontend to __refreshAll """

        self.__refreshAll()

    def setStatusMsg(self, msg):
        """ change status message (i.e. contents of status window) """

        swin = self.__wins["status"]
        realmsg = msg + (" " * (self.__COLS - len(msg)))
            # I'm unable to write the full string, I have to stop at last
            # character or write routines raise "ERR", hence the '- 1'
        swin.addnstr(0, 0, realmsg, self.__COLS - 1)
        swin.refresh()

    def __isCErase(self, char, wname):
        """ return true if char is the 'erase char' for window wname, false
        otherwise """

        if wname == "local":
            return (self.__cEraseL == char)
        elif wname == "remote":
            return (self.__cEraseR == char)
        else:
            raise ValueError(wname)

    def __isWErase(self, char, wname):
        """ return true if char is the 'erase word' for window wname, false
        otherwise """

        if wname == "local":
            return (self.__wEraseL == char)
        elif wname == "remote":
            return (self.__wEraseR == char)
        else:
            raise ValueError(wname)

    def __isLErase(self, char, wname):
        """ return true if char is the 'erase line' for window wname, false
        otherwise """

        if wname == "local":
            return (self.__lEraseL == char)
        elif wname == "remote":
            return (self.__lEraseR == char)
        else:
            raise ValueError(wname)

    def __CErase(self, wname, do_refresh):
        """ delete last inserted char in window 'wname', act like backspace but
        does not work across lines """

        win = self.__wins[wname]
        (y, x) = win.getyx()
        assert x>=0
        if x != 0:  # not on the first column
            win.move(y, x-1)
            win.delch()
        if do_refresh:
            self.__refresh(win)

    def __WErase(self, wname, do_refresh):
        """ delete last inserted word in window 'wname', act like usually
        CTRL-W in shells or emacs (i.e. it stops deleting backward when a blank
        char is encountered; blank chars are those defined in curses.ascii) """

        win = self.__wins[wname]
        (y, x) = win.getyx()
        assert x>=0
        while x>0:  # delete trailing blanks
            win.move(y, x-1)
            char = win.inch()
            win.move(y,x)
            if not curses.ascii.isblank(char):
                break
            else:
                self.__CErase(wname, do_refresh=False)
                (y, x) = win.getyx()
        while x>0:  # delete backward until first blank
            win.move(y, x-1)
            char = win.inch()
            win.move(y, x)
            if curses.ascii.isblank(char):
                break
            else:
                self.__CErase(wname, do_refresh=False)
                (y, x) = win.getyx()    # __CErase move the cursor backward
        if do_refresh:
            self.__refresh(win)

    def __LErase(self, wname, do_refresh):
        """ erase last line inserted in window 'wname', act like CTRL-U in
        shells or emacs """

        win = self.__wins[wname]
        (y, x) = win.getyx()
        assert x>=0
        while x>0:
            self.__CErase(wname, do_refresh=False)
            (y, x) = win.getyx()    # note: __CErase move the cursor backward
        if do_refresh:
            self.__refresh(win)

    def __logEraseChar(self, wname):
        """ frontend to backspace logging, abstract over local and remote
        windows """

        if self.__logger != None:
            if wname == "local":
                self.__logger.logEraseCharL()
            elif wname == "remote":
                self.__logger.logEraseCharR()
            else:
                raise ValueError("Unknown window: '%s'" % wname)

    def __logEraseWord(self, wname):
        """ frontend to erase-word logging, abstract over local and remote
        windows """

        if self.__logger != None:
            if wname == "local":
                self.__logger.logEraseWordL()
            elif wname == "remote":
                self.__logger.logEraseWordR()
            else:
                raise ValueError("Unknown window: '%s'" % wname)

    def __logEraseLine(self, wname):
        """ frontend to erase-line logging, abstract over local and remote
        windows """

        if self.__logger != None:
            if wname == "local":
                self.__logger.logEraseLineL()
            elif wname == "remote":
                self.__logger.logEraseLineR()
            else:
                raise ValueError("Unknown window: '%s'" % wname)

    def __logChar(self, char, wname):
        """ frontend for logging function on local and remote windows,
        recognize end of line characters and log them separately as required by
        logger """

        assert type(char) == int
        if self.__logger != None:
            realchar = chr(curses.ascii.ascii(char))
            assert type(realchar) == str
            assert len(realchar) == 1
            if wname == "local":
                if isEol(realchar):
                    self.__logger.logEolL()
                else:
                    self.__logger.logCharL(realchar)
            elif wname == "remote":
                if isEol(realchar):
                    self.__logger.logEolR()
                else:
                    self.__logger.logCharR(realchar)
            else:
                raise ValueError("Unknown window: '%s'" % wname)

    def __putChar(self, char, src, target, do_refresh):
        """
        lowlevel for putChar{L,R}
            'char' is the char to be displayed
            'src' is the source of the char (can be 'local' or 'remote')
            'target' is the target window (can be 'local' or 'remote')
            'do_refresh' is whether to refresh the target window afterwards
        assumption: type(char) == int
        """

        if self.__isCErase(char, src):    # erase char (i.e. backspace)
            if debug:
                self.__debugger.log("received CErase (%d)" % char)
            self.__CErase(target, do_refresh=False)
            self.__logEraseChar(src)

        elif self.__isWErase(char, src):  # erase word
            if debug:
                self.__debugger.log("received WErase (%d)" % char)
            self.__WErase(target, do_refresh=False)
            self.__logEraseWord(src)

        elif self.__isLErase(char, src):  # erase line
            if debug:
                self.__debugger.log("received LErase (%d)" % char)
            self.__LErase(target, do_refresh=False)
            self.__logEraseLine(src)

        else:                               # normal char
            if debug:
                self.__debugger.log("received normal char (%d)" % char)
            win = self.__wins[target]
            win.addch(char)
            self.__logChar(char, src)

        if do_refresh:  # refresh if needed
            self.__refresh(self.__wins[target])

    def putCharL(self, char, src="local", do_refresh=True):
        """ add a char to local window """

        if type(char) != int:
            raise TypeError("char must be a C-like character (an integer)")
        self.__putChar(char, src, "local", do_refresh)

    def putCharR(self, char, src="remote", do_refresh=True):
        """ add a char to remote window """

        if type(char) != int:
            raise TypeError("char must be a C-like character (an integer)")
        self.__putChar(char, src, "remote", do_refresh)

    def __putString(self, string, wname, do_refresh):
        """
        lowlevel for putString{L,R}
        assumption: type(string) == str
        """

        win = self.__wins[wname]
        for c in string:        # inefficient but avoid duplication of magics
                                # handling code
            self.__putChar(ord(c), wname, False)
        if do_refresh:
            self.__refresh(win)

    def putStringL(self, string, do_refresh=True):
        """ add a string to the local window """

        if type(string) != str:
            raise TypeError("string must have type str")
        self.__putString(string, "local", do_refresh)

    def putStringR(self, string, do_refresh=True):
        """ add a string to the remote window """

        if type(string) != str:
            raise TypeError("string must have type str")
        self.__putString(string, "remote", do_refresh)

    def __patchBS(self, char):
        if char == curses.KEY_BACKSPACE:
            return self.__cEraseL
        else:
            return char

    def getChar(self):
        """ read a char from the console """

        return self.__patchBS(self.__wins["local"].getch())

    def __validateMagicChars(self, magicchars):
        """ validate magic characters argument, raise appropriate exceptions if
        needed """

        if not type(magicchars) == tuple:
            raise TypeError("'magicchars' must have type 'tuple'")
        if not len(magicchars) == 3:
            raise ValueError("'magichars' must be a triple")
        if not ((type(magicchars[0]) == int) and    \
                (type(magicchars[1]) == int) and    \
                (type(magicchars[2]) == int)):
            raise TypeError("'magicchars' must be a int-tuple")

    def setMagicCharsL(self, magicchars):
        """ change local window magic characters """

        self.__validateMagicChars(magicchars)
        (self.__cEraseL, self.__wEraseL, self.__lEraseL) = magicchars
        if debug:
            self.__debugger.log("new local magic chars: <c=%s, w=%s, l=%s>" %\
                            (magicchars[0], magicchars[1], magicchars[2]))

    def getMagicCharsL(self):
        return(self.__cEraseL, self.__wEraseL, self.__lEraseL)

    def setMagicCharsR(self, magicchars):
        """ change remote window magic characters """

        self.__validateMagicChars(magicchars)
        (self.__cEraseR, self.__wEraseR, self.__lEraseR) = magicchars
        if debug:
            self.__debugger.log("new remote magic chars: <c=%s, w=%s, l=%s>" %\
                            (magicchars[0], magicchars[1], magicchars[2]))

    def getMagicCharsR(self):
        return(self.__cEraseR, self.__wEraseR, self.__lEraseR)

    def stop(self):
        """ curses shutdown, must be invoked before exiting to restore screen
        settings """

        curses.nocbreak()
        curses.echo()
        curses.endwin()
        self.__running = False

