# talKZ - (Kalfa and Zack)'s talk implementation
# Copyright (C) 2002:
# 	Cosimo Alfarano      <kalfa@cs.unibo.it>
# 	Stefano Zacchiroli   <zack@cs.unibo.it>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


""" Protocol handling module

    classes to handle protocol stream as automata states.
"""

from select import select
import socket, peer
import messages
from types import *

# TALK_PORT = socket.getsockname("talk", "udp")
NTALK_PORT = socket.getservbyname("ntalk", "udp")
TIMEOUT = 60

voidAddr = messages.Address('0.0.0.0', 0)

debug = False

class Aborted(Exception):
    """ raised when the protocol was aborted for some reason """
    def __init__(self, reason):
        self.__reason = reason
    def getReason(self):
        return self.__reason
    def __str__(self):
        return `self.getReason()`

class Timeout(Exception):
    """ raised when a timeout has been expired """
    def __init__(self, timeout):
        self.__timeout = timeout
    def getTimeout(self):
        return self.__timeout
    def __str__(self):
        return `self.__timeout`

class ClientConnection(object):
    def __init__(self, mymagicchars, capa, ruser, rhost=None, rtty=None):
        if rhost is None:
            rhost = 'localhost'
        if rtty is None:
            rtty = ''
        status = ConnectionStatus(ruser, rhost, rtty)
        status.setMagicChars(mine=mymagicchars)
        status.setCapa(capa)
        self.__state = QuietStatus(status)
    def connect(self, notifier=(lambda _: None)):
        while not isinstance(self.__state, TalkingStatus):
            self.__state.advance()
            state_dsc = self.__state.getDescription()
            if debug:
                print 'description:',self.__state,state_dsc
            if state_dsc != None:
                assert type(state_dsc) == str
                notifier(state_dsc)
            if debug:
                print 'advanced: %s' % str(self.__state.next)
            self.__state = self.__state.next(self.__state.getConnectionStatus())
        if debug:
            print "Hit ENTER to start talking ..."
            import sys
            sys.stdin.readline()
        return self.__state.getTalkingInfo()

class ConnectionAgent(object):
    def __init__(self, rhost=None, rport=NTALK_PORT):
        self.__sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        if rhost: # client
            if debug:
                print 'connecting %s:%s' %(rhost, rport)
            self.__sock.connect((rhost, rport))
            if debug:
                print 'done'
        else: # daemon
            self.__sock.listen(1)
    def getAddress(self):
        """ getAddress(): returns messages.Address(...) instance """
        ip,port = self.getLocalSockName()
        return messages.Address(ip,port)
    def getLocalSockName(self):
        return self.__sock.getsockname()
    def send(self, data):
        return self.__sock.send(data)

    def recv(self, timeout=None):
        size = 1024
        if timeout == None: # not timeout specified, wait forever
            return self.__sock.recv(size)
        else:   # timeout specified, wait max 'timeout' seconds
            if type(timeout) not in [float, int]:
                raise TypeError("'timeout' must be a float or an int")
            else:
                ready_fd = select([self.__sock], [], [], timeout)[0]
                if len(ready_fd) == 0:  # timeout expired
                    raise Timeout(timeout)
                else:   # timeout not expired, data ready
                    return self.__sock.recv(size)

class ConnectionStatus(object):
    def __init__(self, ruser=None, rhost=None, rtty=None): 
        self.__ruser = ruser
        self.__rhost = rhost
        self.__rtty = rtty
        self.__ragent = ConnectionAgent(rhost)
        self.__lagent = ConnectionAgent('localhost')
        self.__talkingSock = socket.socket(socket.AF_INET,
            socket.SOCK_STREAM)
        self.__factory = messages.MessageFactory()
        self.setMagicChars(None,None)
        self.__msgid = {}
        self.setMsgId(0, 'local')
        self.setMsgId(0, 'remote')
        self.setVersion(messages.TALKZ_VERSION)
        self.setAbortReason("") # used by AbortingStatus
        self.setBacklog('')

    def setAbortReason(self, reason):
        if type(reason) != str:
            raise TypeError("'reason' must be a string")
        self.__abort_reason = reason

    def getAbortReason(self):
        return self.__abort_reason

    def getRUser(self):
        return self.__ruser
    def getRTty(self):
        return self.__rtty
    def getRHost(self):
        return self.__rhost
    def getFactory(self):
        return self.__factory
    def getTalkingSock(self):
        return self.__talkingSock
    def getTalkingSockname(self):
        return self.__ragent.getLocalSockName()
    def setTalkingSock(self, host=None):
        if isinstance(host, TupleType):
            self.__talkingSock.connect(host)
        elif isinstance(host, socket.SocketType):
            self.__talkingSock = host
        else:
            self.__talkingSock.listen(1)
        
    def rrecv(self, timeout):
        return self.__ragent.recv(timeout)
    def lrecv(self, timeout):
        return self.__lagent.recv(timeout)
    def recv(self, agent='local', timeout=None):
        if agent == 'local':
            ret = self.lrecv(timeout)
        elif agent == 'remote':
            ret = self.rrecv(timeout)
        else:
            raise ValueError(
                'ConnectionAgent.recv(): agent may be local or ' + \
                'remote, not %s' % agent)
        return ret

    def rsend(self, data):
        return self.__ragent.send(data)
    def lsend(self, data):
        return self.__lagent.send(data)
    def send(self, data, agent='local'):
        if agent == 'local':
            ret = self.lsend(data)
        elif agent == 'remote':
            ret = self.rsend(data)
        else:
            raise ValueError(
                'ConnectionAgent.send(): agent may be local or ' + \
                'remote, not %s' % agent)
        return ret

    def getAddress(self, mode='local'):
        if mode == 'local':
            ret = self.__lagent.getAddress()
        elif mode == 'remote':
            ret = self.__ragent.getAddress()
        else:
            raise ValueError('mode has to be "local" or "remote"')
        return ret

    def getMagicChars(self):
        return self.__mc_mine, self.__mc_its 

    def setMagicChars(self, mine=None, its=None):
        self.__mc_mine = mine
        self.__mc_its = its

    def setMsgId(self, id, mode='remote'):
        self.__msgid[mode] = int(id)

    def getMsgId(self, mode='remote'):
        return self.__msgid[mode]
        
    def getVersion(self):
        return self.__version
    def setVersion(self, version):
        self.__version = version
    def setCapa(self, capa):
        self.__capa = capa
    def getCapa(self):
        return self.__capa
    def mergeCapa(self, capa):
        """ mergeCapa merges the local capability 
        and the one received from peer
        """

        if self.__capa == messages.CAPA['tls']: 
            if capa == messages.CAPA['none']:
                self.__capa = messages.CAPA['none']
            elif not capa == messages.CAPA['tls']:
                raise BadCapability('unkown capability sent by peer')
        elif self.__capa == messages.CAPA['none']:
            pass # no changes needed

        return self.__capa

    def getTLS(self):
        return messages.CAPA['tls'] in self.__capa

    def setBacklog(self, backlog):
        self.__backlog = backlog
    def getBacklog(self):
        return self.__backlog

class ProtocolStatus(object):
    """ Protocol Status Class
        A single status in clientside automata schema
        self.advance() accomplishes current status work
        returns: 
         None : status not fully accomplished
         -1 : error, retrieve information with self.getError() -> string
         ProtocolStatus subclass : OK
    """
    def __init__(self, connstate):
        self.next = None
        self.__history = []
        self.__connstate = connstate
        self.__error = None
        self.__dsc = None

    def _appendHistory(self):
        self.__history.append(self.__class__)

    def getHistory(self):
        return self.__history

    def getConnectionStatus(self):
        return self.__connstate

    def getDescription(self):
        return self.__dsc
    def setDescription(self, dsc):
        self.__dsc = dsc

    def advance(self):
        """ Accomplishes current status operations, advancing over Protocol Automata
            Returns next Protocol Class and sets self.next to it.
        """
        raise NotImplementedError('advance()')

    def sendDeleteS(self):
        """ send DELETE requests to both local and remote servers """

        factory = self.getConnectionStatus().getFactory()
        rname = self.getConnectionStatus().getRUser()
        rtty = self.getConnectionStatus().getRTty()
        version = self.getConnectionStatus().getVersion()
        for dest in ['remote','local']:
            ctladdr = self.getConnectionStatus().getAddress(dest)
            id = self.getConnectionStatus().getMsgId(dest)
            req = factory.buildRequest(
                addr=voidAddr, ctl_addr=ctladdr, id_num=id, r_name=rname,
                r_tty=rtty, vers=version, typ=messages.DELETE)
            data = req.pack()
            self.getConnectionStatus().send(data, dest)

    def versionFallBack(self):
        status = self.getConnectionStatus()
        version = status.getVersion()
        print 'version fallback',status.getVersion(),
        if version == messages.TALKZ_VERSION:
            status.setVersion(messages.TALK_VERSION)
        else:
            raise classes.BadVersionError("Remote Server doesn't support any of my current supported talk version") 
        print status.getVersion()
        
class QuietStatus(ProtocolStatus):
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)
    def advance(self):
        ret = LookUpStatus
        status = self.getConnectionStatus()
        factory = status.getFactory()
        addr = voidAddr
        ctladdr = self.getConnectionStatus().getAddress('remote')
        rname, rtty = status.getRUser(), status.getRTty()
        version = self.getConnectionStatus().getVersion()
        req = factory.buildRequest(addr=voidAddr, ctl_addr=ctladdr,
            r_name=rname, r_tty=rtty,
            vers=version, typ=messages.LOOK_UP)
        data = req.pack()

        self.getConnectionStatus().send(data, 'remote')
        self._appendHistory()

        if ret:
            self.next = ret
        return ret

class LookUpStatus(ProtocolStatus):
    """ CallerSide ans CalleeSide Status
        Asks for calle or caller presence to daemon
        I cannot know if I'm a caller or callee: depending on how
        daemon answer I will connect to caller or ask for a remote
        callee

        [Operator?]
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)
    def advance(self):
        ret = SendAnnounceStatus # by default I believe to be a caller
        
        status = self.getConnectionStatus()
        allowedAnswers = [ messages.NOT_HERE, messages.SUCCESS, \
            messages.BADVERSION ]

        self.setDescription('Looking up callee')

        factory = self.getConnectionStatus().getFactory()
        data = status.recv('remote')
        resp = factory.buildResponseFromBinary(data)
       
        if int(resp.getAnswer()) not in allowedAnswers:
            status.setAbortReason( "Unexpected answer: %s" % \
                messages.string_of_answer(resp.getAnswer()))
            ret = AbortingStatus
        
        status.setMsgId(resp.getId(), 'remote')

        if resp.getAnswer() == messages.BADVERSION:
            self.setDescription('Falling back to prev version')
            ret = BadVersionStatus
        elif resp.getAnswer() == messages.SUCCESS:
            # we're a callee, now we know it!
            self.setDescription('Looking up caller')
            ret = ConnectedActiveStatus

            status.setVersion(resp.getVersion())

            raddr = resp.getAddr().getStrAddr(), resp.getAddr().getPort()
            try:
                self.getConnectionStatus().setTalkingSock(raddr)
            except socket.error, (errno, _):
                if errno == 111:    # 'connection refused' error
                    status.setAbortReason("Connection refused by peer")
                    ret = AbortingStatus
                else:
                    raise

        self._appendHistory()
        if ret:
            self.next = ret
        return ret

class BadVersionStatus(ProtocolStatus):
    """ CallerSide ans CalleeSide Status
        We've sent a unsupported version to daemon.
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)
    def advance(self):
        ret = QuietStatus # by default I believe to be a caller
        if self.getConnectionStatus().getVersion() == \
                messages.TALKZ_VERSION:
            self.getConnectionStatus().setVersion(messages.TALK_VERSION)
        else:
            raise classes.BadVersionError(
                "cannot handle any daemon supported version")
        self.next = ret
        return ret

class SendAnnounceStatus(ProtocolStatus):
    """ CallerSide Status
        Contact Remote Daemon sending ANNOUNCE about the will of talking
        of current client
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)
    def advance(self):
        ret = RecvAnnounceStatus
        
        status = self.getConnectionStatus()
        if debug:
            print 'SendAnnounceStatus' 
            print 'Version is', status.getVersion()
        factory = status.getFactory()
        ip,port = status.getTalkingSock().getsockname() 
        if debug:
            print 'SendAnnounceStatus: TCPSock: %s:%d' % \
                (status.getTalkingSockname(), port) 
        addr = messages.Address(status.getTalkingSockname()[0], port)
        ctladdr = status.getAddress('remote')
        version = status.getVersion()
        if debug:
            print 'SendAnnounceStatus: CTLAddr: %s:%d' % \
            (ctladdr.getStrAddr(), ctladdr.getPort())
        rname = status.getRUser()
        rtty = status.getRTty()
        req = factory.buildRequest(addr=addr, ctl_addr=ctladdr,
            r_name=rname, r_tty=rtty, 
            vers=version, typ=messages.ANNOUNCE)
        data = req.pack()

        self.getConnectionStatus().send(data, 'remote')
        if ret:
            self.next = ret
        return ret

class RecvAnnounceStatus(ProtocolStatus):
    """ CallerSide Status
        Contact Local Daemon sending LEAVE_INVITE about the will of talking
        of current client
        
    """

    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)

    def advance(self):
        ret = SendLeaveInviteStatus
        
        allowedAnswers = [ messages.SUCCESS, messages.BADVERSION ]
        factory = self.getConnectionStatus().getFactory()

        status = self.getConnectionStatus()
        try:
            data = status.recv('remote', TIMEOUT)
            timeout = False
        except Timeout:
            timeout = True
        if not timeout: # timeout not expired
            resp = factory.buildResponseFromBinary(data)
            status.setMsgId( resp.getId(), 'remote')
            if int(resp.getAnswer()) not in allowedAnswers:
                if resp.getAnswer() == messages.NOT_HERE:
                    status.setAbortReason("User not logged in")
                elif resp.getAnswer() == messages.PERMISSION_DENIED:
                    status.setAbortReason("User is refusing messages") 
                else:
                    status.setAbortReason("Unexpected answer: %s" % \
                            messages.string_of_answer(resp.getAnswer()))
                ret = AbortingStatus
            elif resp.getAnswer() == messages.BADVERSION:
                self.versionFallBack()
                self.setDescription('version falling back')
                ret = SendAnnounceStatus
        else:   # timeout expired
            self.setDescription(
                'timeout expired waitin for LEAVE_INVITE success')
            ret = SendAnnounceStatus

        if ret:
            self.next = ret
        return ret

class SendLeaveInviteStatus(ProtocolStatus):
    """ CallerSide Status
        Contact Local Daemon sending LEAVE_INVITE about the will of talking
        of current client

        [Voulez-vous causer avec moi?]
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)
    def advance(self):
        ret = RecvLeaveInviteStatus
       
        status = self.getConnectionStatus()
        
        status.setTalkingSock()
        factory = status.getFactory()
        ip,port = status.getTalkingSock().getsockname() 
        if debug:
            print 'StartingStatus: TCPSock: %s:%d' % \
                (status.getTalkingSockname(), port) 
        addr = messages.Address(status.getTalkingSockname()[0], port)
        ctladdr = status.getAddress('local')
        rname = status.getRUser()
        rtty = status.getRTty()
        version = status.getVersion()

        req = factory.buildRequest(addr=addr, ctl_addr=ctladdr,
            r_name=rname, r_tty=rtty, 
            vers=version, typ=messages.LEAVE_INVITE)
        data = req.pack()

        status.send(data, 'local')
        if ret:
            self.next = ret
        return ret

class RecvLeaveInviteStatus(ProtocolStatus):
    """ CallerSide Status
        Contact Local Daemon sending LEAVE_INVITE about the will of talking
        of current client
        
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)

    def advance(self):
        ret = WaitingConnectionStatus
        
        allowedAnswers = [ messages.SUCCESS, messages.BADVERSION ]
        factory = self.getConnectionStatus().getFactory()

        status = self.getConnectionStatus()
        try:
            data = status.recv('local', TIMEOUT)
            timeout = False
        except Timeout:
            timeout = True
        if not timeout: # timeout not expired
            resp = factory.buildResponseFromBinary(data)
            status.setMsgId( resp.getId(), 'local')
            if int(resp.getAnswer()) not in allowedAnswers:
                if resp.getAnswer() == messages.NOT_HERE:
                    status.setAbortReason("User not logged in")
                elif resp.getAnswer() == messages.PERMISSION_DENIED:
                    status.setAbortReason("User is refusing messages")
                else:
                    status.setAbortReason("Unexpected answer: %s" % \
                            messages.string_of_answer(resp.getAnswer()))
                ret = AbortingStatus
            elif resp.getAnswer() == messages.BADVERSION:
                self.versionFallBack()
                self.setDescription('version falling back')
                ret = SendLeaveInviteStatus
        else:   # timeout expired
            self.setDescription(
                'timeout expired waitin for LEAVE_INVITE success')
            ret = SendLeaveInviteStatus

        if ret:
            self.next = ret
        return ret

class AbortingStatus(ProtocolStatus):
    """ abort for whatever reason; reason is retrieved from connstate """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)
        self.sendDeleteS()
        raise Aborted(connstate.getAbortReason())

class WaitingConnectionStatus(ProtocolStatus):
    """ CallerSide Status
        Wait for Peer TCP Connection
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)

    def advance(self):
        ret = ConnectedPassiveStatus
        
        acceptedsock = self.getConnectionStatus().getTalkingSock().accept()
        self.sendDeleteS()
        self.getConnectionStatus().setTalkingSock(acceptedsock[0])

        if ret:
            self.next = ret
        return ret
        
class ConnectedPassiveStatus(ProtocolStatus):
    """ CallerSide and CalleeSide Status
        Send MagicChars for remove char, word and line.

        [Supercalifragilisticespiralidoso]
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)
    def advance(self):
        ret = ConnectionPassiveStatus
        
        status = self.getConnectionStatus()
        talksock = status.getTalkingSock()
        mymc, itsmc = status.getMagicChars()
        talksock.send(reduce(lambda x,y: x+chr(y), mymc, ""))
        itsmc = ""

        while len(itsmc) != 3:
            data = talksock.recv(3-len(itsmc))
            itsmc += data
        assert len(itsmc) == 3
        status.setMagicChars(its=(itsmc[0], itsmc[1], itsmc[2]))

#         talksock.send('ciao\n')
#         print talksock.recv(1024)
#         talksock.send('ciao2\n')
#         print talksock.recv(1024)
        
        self.next = ret
        return ret

class ConnectionPassiveStatus(ProtocolStatus):
    """ CallerSide Status
        Choose for Traditional or Enhanced protocol and computes
        capability in the latter case.
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)

    def advance(self):
        ret = TalkingStatus

        status = self.getConnectionStatus()
       
        sock = status.getTalkingSock()
        ready_fd = select([sock], [], [], 1)[0]
        if sock in ready_fd:
            data = sock.recv(messages.CAPA_LEN) 
            if data in messages.CAPA.values():
                status.mergeCapa(data)
                ret = CapabilitiesRecvdStatus
            else:
                status.setBacklog(data)

        if ret:
            self.next = ret

        return ret

class CapabilitiesRecvdStatus(ProtocolStatus):
    """ CallerSide Status 
        Sends computed capability
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)
    def advance(self):
        ret = TalkingStatus
        
        status = self.getConnectionStatus()
        status.getTalkingSock().send(status.getCapa())

        status = self.getConnectionStatus()

        if ret:
            self.next = ret
        return ret
    
class ConnectedActiveStatus(ProtocolStatus):
    """ CallerSide and CalleeSide Status
        Send MagicChars for remove char, word and line.

        [Supercalifragilisticespiralidoso]
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)
    def advance(self):
        ret = ConnectionActiveStatus
        
        status = self.getConnectionStatus()
        talksock = status.getTalkingSock()
        mymc, itsmc = status.getMagicChars()
        talksock.send(reduce(lambda x,y: x+chr(y), mymc, ""))
        itsmc = ""

        while len(itsmc) != 3:
            data = talksock.recv(3-len(itsmc))
            itsmc += data
        assert len(itsmc) == 3
        status.setMagicChars(its=(itsmc[0], itsmc[1], itsmc[2]))

#         talksock.send('ciao\n')
#         print talksock.recv(1024)
#         talksock.send('ciao2\n')
#         print talksock.recv(1024)
        
        self.next = ret
        return ret

class ConnectionActiveStatus(ProtocolStatus):
    """ CalleeSide Status
        
        At the party gate, callee have to choose to use enhanced
        or traditional protocol.

        [To be or not to be?]
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)
 
    def advance(self):
        ret = TalkingStatus

        if self.getConnectionStatus().getVersion() == \
                messages.TALKZ_VERSION:
            ret = CapabilitiesSentStatus
           
            assert len(self.getConnectionStatus().getCapa()) == 1
            self.setDescription("enhanced version")
            self.getConnectionStatus().getTalkingSock().send( \
                self.getConnectionStatus().getCapa() )
        else:
            self.setDescription("traditional talk version")
            status = self.getConnectionStatus()
 
        if ret:
            self.next = ret
        return ret

class CapabilitiesSentStatus(ProtocolStatus):
    """ CalleeSide Status
    """
    def __init__(self, connstate):
        ProtocolStatus.__init__(self, connstate)
    def advance(self):
        ret = TalkingStatus
        
        sock = self.getConnectionStatus().getTalkingSock()
        capa = sock.recv(messages.CAPA_LEN)
        
        self.getConnectionStatus().setCapa(capa)

        if messages.CAPA['tls'] in capa:
            string = 'TLS enabled'
            self.setDescription('received capabilities: %s' % string)

        status = self.getConnectionStatus()

        if ret:
            self.next = ret
        return ret

class TalkingStatus(ProtocolStatus):
    """ CallerSide and CalleeSide Status
        Talk with your partner, guy!
    """
    def getTalkingInfo(self):
        status = self.getConnectionStatus()
        p = peer.Peer(status.getTalkingSock(), status.getTLS())
        return  \
            (map(ord, status.getMagicChars()[1]), 
                p, status.getBacklog())

#c = ClientConnection('123', 'kalfa', 'localhost', '')
#c.connect()
