#!/usr/bin/python

# talKZ - (Kalfa and Zack)'s talk implementation
# Copyright (C) 2002:
# 	Cosimo Alfarano      <kalfa@cs.unibo.it>
# 	Stefano Zacchiroli   <zack@cs.unibo.it>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import signal, socket, sys, syslog, time
import announce, invites, messages, kzstruct


MAXIDLE = 120   # maximum number of seconds which the daemon can remain idle
                # (i.e. no message received), after this value the daemon will
                # exit
                # value from "talkd.c" of netkit-ntalk


def udpSend(data, dest, sock):
    """ send 'data' to 'dest' using udp socket 'sock'. Return only when all
    data have been sent. 'data' must be a string, 'dest' a pair <ipaddr, port>,
    'sock' an UDP socket """

    residual = data
    while len(residual) > 0:
        sent = sock.sendto(residual, dest)
        assert sent <= len(residual)
        if sent > 0:
            residual = residual[:-sent]


def doAnnounce(req, src_hostname):
    """
    Display a talk request message to one or more callee user's ttys.
    Return an answer value suitable for "answer" field of a Response object;
    possible values are:
        messages.FAILED     user is refusing messages on all its ttys
        messages.SUCCESS    at least one tty was notified
        messages.NOT_HERE   user not logged in
    """

    msg = "Message from Talk_Daemon@%s at %s ...\n"
    msg += "talk: connection requested by %s@%s.\n"
    msg += "talk: respond with:\ttalk %s@%s\n"
    msg = msg % (socket.gethostname(), time.ctime(),
                 req.getCallerName(), src_hostname,
                 req.getCallerName(), src_hostname)
    user = req.getCalleeName()
    tty = req.getCalleeTty()
    try:
        ttys_ringed = announce.announce(user, msg, tty)
        if ttys_ringed == 0:    # user is refusing messages
            return messages.FAILED
        else:   # ttys_ringed > 0, at least one tty was notified
            return messages.SUCCESS
    except announce.NotLoggedIn: # user not logged in
        return messages.NOT_HERE


class TalKZDConf(object):
    """
    talKZ Daemon configuration.

    This object is intended to remember start up settings of the daemon; other
    settings which are changed at runtime _are_not_ kept here.

    To obtain a default configuration object just instantiate TalKZDConf with
    no arguments.

    Methods:
        setDebug(bool)          enable/disable debugging
        getDebug()              return current debugging setting
        setProtocol(protocol)   set the default message protocol
        getProtocol()           return che current message protocol
        setStandalone(value)    set whether talkzd is a standalone daemon
                                (value == True) or a inetd daemon
                                (value == False)
        isStandalone()          return true if talkzd is standalone
    """

    def __init__(self, debug=False, protocol="bsd", standalone=False):
        self.setDebug(debug)
        self.setProtocol(protocol)
        self.setStandalone(standalone)

    def setStandalone(self, value):
        if (value != True) and (value != False):
            raise ValueError("'value' must be a boolean")
        self.__standalone = value

    def isStandalone(self):
        return (self.__standalone == True)

    def setDebug(self, debug):
        if (debug != True) and (debug != False):
            raise ValueError("'debug' must be a boolean")
        self.__debug = debug

    def getDebug(self):
        return self.__debug

    def setProtocol(self, proto):
        if (proto != "sunos") and (proto != "bsd"):
            raise ValueError("'proto' must be one of 'bsd', 'sunos'")
        self.__protocol = proto

    def getProtocol(self):
        return self.__protocol


def timeout(signum, frame):
    assert signum == signal.SIGALRM
    sys.exit(0)


class InvalidRequest(Exception):
    """ raised after receiving an invalid request """

    def __init__(self, req, msg):
        """
        InvalidRequest(msg, req)    where 'msg' is an error message describing
                                    the problem and 'req' is the problematic
                                    request object
        """

        self.__req = req
        self.__msg = msg

    def __str__(self):
        return `self.__msg`

    def getReq(self):
        return self.__req

    def getMsg(self):
        return self.__msg


class TalKZD(object):

    def __init__(self, conf=TalKZDConf()):

        self.__conf = conf
        self.__factory = messages.MessageFactory(self.__conf.getProtocol())
        self.__invites = invites.InvitesTable()
        if self.__conf.isStandalone():  # standalone daemon
                                        # (meant only for debugging and
                                        # launched interactively!!!)
            self.__insock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.__insock.bind(("0.0.0.0", 518))
            self.__outsock = self.__insock
        else:   # inetd dependent daemon
            self.__insock = socket.fromfd(sys.stdin.fileno(), \
                                    socket.AF_INET, socket.SOCK_DGRAM)
            self.__outsock = socket.fromfd(sys.stdout.fileno(), \
                                    socket.AF_INET, socket.SOCK_DGRAM)
        try:
            self.__insock.getsockname()
            self.__outsock.getsockname()
        except socket.error, (errno, _): # catch 'socket op on a non socket fd'
            if errno == 88:
                sys.stderr.write("stdin and/or stdout aren't UDP sockets, " + \
                                 "please run talKZd through inetd\n")
                sys.exit(1)
        except: # reraise other exceptions
            raise
        syslog.openlog("talkzd", syslog.LOG_PID, syslog.LOG_DAEMON)
        self.__debug("debug", "Starting up ...")
        self.__startTimer()
        self.__mainloop()

    def __startTimer(self):
        signal.signal(signal.SIGALRM, timeout)
        signal.alarm(MAXIDLE)

    def __restartTimer(self):
        signal.alarm(MAXIDLE)

    def __del__(self):
        self.__debug("debug", "Exiting ...")
        syslog.closelog()

    def __debug(self, typ, msg):
        """ Debugging logging. Log messages passed through this function are
        logged using syslog only if debugging is enabled (i.e.
        self.__conf.getDebug == True), priority is determined depending on
        value of 'typ' that can be "info" or "debug" """

        if self.__conf.getDebug():
            if typ == "debug":
                if self.__conf.isStandalone():  # standalone
                    print "DEBUG: %s" % msg
                else:   # through inetd
                    syslog.syslog(syslog.LOG_DEBUG, msg)
            elif typ == "info":
                if self.__conf.isStandalone():  # standalone
                    print "INFO: %s" % msg
                else:   # through inetd
                    syslog.syslog(syslog.LOG_INFO, msg)
            else:
                raise ValueError("'typ' must be one of 'debug', 'info'")
            sys.stdout.flush()

    def __log(self, msg):
        """ Like __debug but meant for more important message, not only related
        to debugging like classical syslog daemon messages """

        if self.__conf.isStandalone():  # standalone
            print "DAEMON: %s" % msg
        else:   # through inetd
            syslog.syslog(msg)

    def __receiveReq(self):
        """ Read a request from the control socket, perform some sanity checks,
        unpack the request creating a Request instance.
        Return a pair <req, sender> where req is the Request object and sender
        is a pair <addr, port> where address is the source IP address (in
        string) and port is the source port address of the request """

        reqsize = self.__factory.getProtocolClass().getReqSize()
        bufsiz =  reqsize * 2
        (rawdata, (addr, port)) = self.__insock.recvfrom(bufsiz)
        self.__log("Connection from: %s" % `(addr, port)`)
        self.__restartTimer()   # restart timer alway
        self.__debug("debug",
                "request received from %s, port %d, size: %d byte(s)" % \
                    (addr, port, len(rawdata)))
        if len(rawdata) != reqsize:    # size mismatch
            raise InvalidRequest(None,
                "size mismatch: %d byte(s) received, %d byte(s) expected" % \
                    (len(rawdata), reqsize))
        try:
            req = self.__factory.buildRequestFromBinary(rawdata)
        except kzstruct.Error:  # trash received
            raise InvalidRequest(None, "request unpacking failed")
        return (req, (addr, port))

    def __processReq(self, req, (src_addr, src_port)):
        self.__debug("debug", "Processing request:\n" + `req`)
        res = self.__factory.buildResponse( \
            vers=req.getVersion(), typ=req.getType(), id_num=0)
        if req.getVersion() not in messages.VERSIONS:   # bad version
            self.__log("Bad protocol version: %d" % req.getVersion())
            res.setAnswer(messages.BADVERSION)
        else:   # version ok
            reqType = req.getType()

            if reqType == messages.LEAVE_INVITE:    # LEAVE_INVITE req
                self.__debug("debug", "request is a LEAVE_INVITE one")
                try:
                    prev_req = self.__invites.findRequest(req)
                    self.__debug("debug", "invite already left")
                    res.setId(prev_req.getId())
                    res.setAnswer(messages.SUCCESS)
                except LookupError: # invite not found
                    newid = self.__invites.add(req) # may change req._id_num
                    self.__debug("debug", "added invite with id %d" % newid)
                    res.setId(req.getId())

            elif reqType == messages.LOOK_UP:       # LOOK_UP req
                self.__debug("debug", "request is a LOOK_UP one")
                try:
                    partner_req = self.__invites.findPartner(req)
                        # partner found!
                    self.__debug("debug", "partner found")
                    res.setId(partner_req.getId())
                    if req.getVersion() == messages.TALKZ_VERSION: # talkz peer
                            # override version which caller's one
                        res.setVersion(partner_req.getVersion())
                    res.setAddr(partner_req.getAddr())
                    res.setAnswer(messages.SUCCESS)
                except LookupError: # partner not found
                    self.__debug("debug", "partner not found")
                    res.setAnswer(messages.NOT_HERE)

            elif reqType == messages.DELETE:        # DELETE req
                self.__debug("debug", "request is a DELETE one")
                try:
                    self.__invites.delete(req.getId())
                    res.setAnswer(messages.SUCCESS)
                    self.__debug("debug", "deleting id %s" % \
                        str(req.getId()))
                except LookupError: # invite not found
                    res.setAnswer(messages.NOT_HERE)

            elif reqType == messages.ANNOUNCE:      # ANNOUNCE req
                self.__debug("debug", "request is an ANNOUNCE one")
                try:
                    src_hostname = socket.gethostbyaddr(src_addr)[0]
                except socket.herror, (errno, _):   # no reverse for src_addr
                    src_hostname = src_addr # use ip address
                try:
                    prev_req = self.__invites.findRequest(req)
                    if req.getId() > prev_req.getId():  # explicit reannounce
                        prev_req.setId(self.__invites.newId())  # update cached
                                                                # id
                        res.setId(prev_req.getId())
                        res.setAnswer(doAnnounce(req, src_hostname))
                    else:   # duplicate announce
                        res.setId(prev_req.getId())
                        res.setAnswer(messages.SUCCESS)
                except LookupError: # first announce
                    self.__invites.add(req) # Note: may modify req._id_num
                    res.setId(req.getId())
                    res.setAnswer(doAnnounce(req, src_hostname))

            else:   # bad request
                self.__log("Unknwon request: %d" % reqType)
                res.setAnswer(messages.UNKNOWN_REQUEST)

        return res

    def __sendRes(self, res, (dest_addr, dest_port)):
        """ send response 'res' to destination 'dest'; 'res' must be a Response
        object, 'dest' must be a pair <ipaddr, port> """

        if not isinstance(res, messages.Response):
            raise TypeError("'res' must be a Response object")
        self.__debug("debug", "Destination: <%s, %d>" % (dest_addr, dest_port))
        self.__debug("debug", "Sending response:\n" + `res`)
        udpSend(res.pack(), (dest_addr, dest_port), self.__outsock)

    def __mainloop(self):
        while True:
            try:
                (req, sender) = self.__receiveReq()
                res = self.__processReq(req, sender)
                self.__sendRes(res, sender)
            except InvalidRequest, exc:
                self.__debug("info",
                        "Invalid request received: %s" % exc.getMsg())
                        
                continue


def main():
    conf = TalKZDConf()
    # conf.setStandalone(True)  # default is through inetd
    conf.setDebug(True) # default is no debugging
    TalKZD(conf)


if __name__ == '__main__':
    main()

